#!/usr/bin/python2
"""\
forge is a python based build mechanism that is partially inspired by
build.exe (www.imonk.com) and partially inspired by scons (www.scons.org).

All of forge is implemented in the single forge.py file, which is to say
both the executable and the module.

The primary goals of forge are:
- easy to write build description files
- portability
- powerful, as in the power of a full language at hand
- simple building blocks without a lot of usage restrictions or assumptions

Execution Environment and Configuration Files
=============================================
To build a configured project using forge, run forge.py from any directory
within the project.  Forge searches up the directory path to find the
root.py file.  The root.py file is the main project file, so there should
be only one root.py per project.  Use the -f flag to override the name
root.py.

The root.py may import modules with the setup() method which does a
specialized python 'import' to bring in build modules.	The setup function
will look for a 'bld.py' in the specified module.  The module subdirectory
is derived from the module name (and visa versa such as when modules are
imported from other modules).  The method setup(module) within the bld.py
file is called upon import.

To access forge as a python module from a build module, do so by finding it
in the sys provided modules dictionary:
forge = sys.modules["forge"]

Primary targets are specified on the command line as TargetRegistry entry
names (see Target Registry below).

To check if a target is up-to-date it must be at least as new as some
'reference time'.  This reference time is specifed to the build function,
but is also modified by dependencies.  Therefore, the reference time will
be the newest of all dependencies and the specified reference time.

The default specified reference time to primary targets on the command line
is 0 (in other words, very long ago).  Therefore, the default build mode
is to make sure primary targets are as new as their dependencies only.
The -x flag sets the specified reference time to the current time which
will have the effect of rebuilding almost everything.  This is often
preferable to an explicitly driven full rebuild via `forge.py clean root`,
since 'clean' is more prone to missing something (because forge has no
way of implicitly knowing what all transient targets are; clean uses
pattern matching to get what it gets).

Platform Abstraction Variables
==============================
Core forge feature platform abstraction is provided through variables.
These are setup by the initialize() function.  Currently, only Win32/VC++
and Linuc/g++ are built-in.  To provide a new platform simply reassign
these variables.

Targets and the Dependency Graph
================================
Targets are python objects representing some entity that has the concept
of being 'up to date'.	Typically, a target represents a file.

Dependencies represent the relationship between targets that if one target
is dependent on another target it must be at least as new as that target.

Currently, the dependency graph is built from scratch every time forge
is executed.  So far this has not been a significant performance issue.
(If it becomes an issue, graph caching to file is a reasonably easy fix.  This
is in effect what typical 'make dep' does by appending to makefiles.).

Dependencies are added with Target::AddDep(), Target::AddDeps(), dep(),
or deps().	dep() and deps() now do the same thing, which they now both
handle list AND non-list arguments.  They both also handle Target objects
in additional to the previously supported string arguments.

Configuration files may add dependencies explicitly.  Also, dependencies
are added automatically via several mechanisms:
- Included C/C++ file scanner; recursively scans header dependencies.
- Build file automatic additions
Of course, automatic dependency addition is not restricted to included
methods.

Target Registry and Target Dictionary
-------------------------------------
All targets are also inserted into the target dictionary, keyed off
of their target names.	This dictionary is used for global operations,
such as the C/C++ header dependency scanner.

Targets may also be added to the target registry which is not necessarily
keyed off of target names.	Forge employs naming conventions on the
registry to facilitate useful cross-module dependency specification.
Some conventions:
- ROOT_BUILD_FILE: the main root.py file
- <module name>.buildfile: a particular module's build file
- <module name>: the module target
- <Module::Exe() specified name>Exe: executable target
- <Module::Lib() specified name>Lib: library target
- <Module::DLL() specified name>Lib: library target
- clean: clean target
- print: print target
- root: root module target

Note that Libs and DLLs collide in Registry namespace.	This is for several
reasons.  First, cross-module dependencies are easier to specify in an
project where Libs and DLLs may be switched.  Second, in Win32 there is
an IMPLIB 'Lib' for every 'DLL', so having a true static 'Lib' and a 'DLL' at
the same time is problematic.  This is of course all overridable.

The target registry is also used to look up primary targets from the
command line.  So to build just a particular executable you may run:
forge.py particularExe

Rules and Resolving Dependencies
================================
A rule is a regular expresion and an associated python function with the
prototype:
def rule(target)

Rule Contexts
-------------
A rule context is a collection of rules.  The three significant contexts
are the pre-processing context (preContext), the build
context (depContext), and the post-processing context (postContext).
The depContext can be overridden on a per target basis by simply assigning
the data member 'depContext' of a target.

The default depContext (defaultContext) currently includes some
reasonable default rules for Win32 and Linux C/C++ projects.

A rule is executed against a target if the target name is matched
by the regular expression of the rule.	When the matching is checked
depends upon the context (see Target::Build()).

For example, to add a rule to print a message every time a win32 or linux
object file is built you might have:
def print_obj(target):
	print 'building: ' + target.name
forge.defaultContext.AppendRule(r"(.*)\.(obj|o)", print_obj)

Target::Build(reference_time), the heart of forge
-------------------------------------------------
(see Target::Build())

Some Useful Forge Variables
===========================

executable command line maps
----------------------------
These dictionaries are used to assemble command lines for compiling and
linking.  These are global and per target dictionaries.  A module
dictionary entry overrides a global dictionary entry.  All existing keys are
used.

includemap -  Header file include directories
			  includeScanSuppress is a list of includemap keys to suppress
			  the dependency scan for.	This is useful to remove the overhead
			  of maintaining dependencies on large unchanging third party
			  packages.
cppmap -	  Compiler command line flags
linkmap -	  Linker command line flags

others
------
preContext -  The pre-processing rule context
depContext -  The main build rule context.
postContext - The post-processing rule context.
rootPath -	  The absolute path to the project root directory.
libPath -	  Used by the module convienience methods, Exe, Lib, and DLL to
			  place the targets.  Also added to library search path.
objPath -	  Used by automatic object file target generation function,
			  ObjAndSrcTargets(), to create object file targets.  Note that
			  the relative path structure is preserved (so for objPath=obj
			  and target root/module/a.cpp, target root/obj/module/a.o will
			  be added)
binPath -	  Not used by forge directly; potentially used by root.py.


root.py guide
=============

bld.py guide
============

A Super Simple Example
======================

root build file (root.py)
-------------------------
import sys
forge = sys.modules["forge"]
forge.rootModule.Exe("HelloWorld")


A More Interesting Example
==========================

the desired dependency graph
----------------------------

				root
				/|\ 
		 ------- | --------
		/		 |		   \\
  moduleLib  common.h	 moduleExe
   /	\	/		\	  /    \\
  /		 \ /		 \	 /		\\
a.cc -> a.h <------- b.h <--- b.cc

root build file (root.py)
-------------------------
import sys

forge = sys.modules["forge"]

forge.includemap['common_include'] = '.'
forge.includemap['lib_include'] = 'moduleLib'

module_list = [ 'moduleLib',
				'moduleExe' ]

for mod in module_list:
	forge.setup(mod)

moduleLib build file (moduleLib/bld.py)
---------------------------------------
import sys

forge = sys.modules["forge"]

def setup(module):
	module.Lib("My", ["a"])

moduleExe build file (moduleExe/bld.py)
---------------------------------------
import sys

forge = sys.modules["forge"]

def setup(module):
	exe = module.Exe("My", ["b"])
	forge.dep("MyExe", "MyLib")


"""



















































__author__ = "Andrew J. Weber <orang@imonk.com>"
__date__ = "3 September 2003"

ccache=0
distcc=0

###############################################################################
# AS THE DRIVER
###############################################################################
if __name__ == '__main__':
	import os
	import sys
	import time
	import forge

	forge.reset_buildtime()

	forge.initialize()

	# find a root.py
	#rootPath = os.getcwd()
	rootPath = forge.rootPath
	lastPath = ""
	haveRoot = 0
	while rootPath != lastPath:
		lastPath = rootPath
		if os.path.isfile( os.path.join(rootPath,forge.rootFile)):
			forge.SetRootPath(rootPath)
			#root = forge.import_from('root',rootPath)
			haveRoot = 1
			break
		(rootPath,drop) = os.path.split(rootPath)

	# make deferral convienience
	if not haveRoot:
		forge.cprint(forge.RED,1,'no root.py found')
		if os.path.isfile( 'makefile' ) or os.path.isfile( 'Makefile' ):
			forge.cprint(forge.RED,1,'found makefile, deferring to make')
			command = 'make'
			for s in sys.argv[1:]:
				if s == 'root':
					s = 'all'
				if s == '-x':
					s = 'clean'
				command = command + ' ' + s
			os.system(command)
			sys.exit(0)

	# if root.py defers to another root
	if forge.path_to_root != "":
		rootPath = os.path.join(os.getcwd(),forge.path_to_root)
		rootPath = os.path.abspath(rootPath)
		forge.SetRootPath(rootPath)

	# import configuration
	config_name_list = ""
	for c in forge.configModules:
		try:
			forge.import_from(c, forge.rootPath)
		except:
			pass


	bld = forge.import_from('root', rootPath)


	# post hook mechanism
	for (k,v) in forge.targetDictionary.items():
		if v.post_hook != None:
			v.post_hook(v)

	#forge.targetRegistry['ROOT_BUILD_FILE'] = forge.FileTarget(
	#	os.path.join(forge.rootPath,forge.rootFile))
	#forge.args = forge.HandleCommandLine(sys.argv)

	#if not forge.rootPath:
	#	forge.SetRootPath(os.getcwd())

	if forge.ccache != 0:
		ccache_dir=os.getenv('CCACHE_DIR')
		if(ccache_dir==None):
			forge.cprint(forge.CYAN,0,'ccache: using default')
		else:
			forge.cprint(forge.CYAN,0,'ccache: '+ccache_dir)
	else:
		forge.cprint(forge.YELLOW,0,'ccache: not installed')

	if forge.distcc != 0:
		forge.cprint(forge.CYAN,0,'distcc: found')
	else:
		forge.cprint(forge.YELLOW,0,
				'distcc: not installed or DISTCC_HOSTS not set')

	forge.cprint(forge.CYAN,0,'rootPath: ' + forge.rootPath)
	forge.cprint(forge.CYAN,0,'api: ' + forge.api)

	forge.ScanAllTargets()

	if not forge.failed:
		for arg in forge.args:
			match = forge.MatchTarget(arg)
			forge.buildNum += 1
			if match != None:
				match.CycleCheck()
				match.Build(forge.referenceTime)
			else:
				forge.cprint(forge.RED,0,'WARNING: unknown target: ' + arg)


	if forge.printCommon != 0:
		keylist = forge.targetRegistry.keys()
		keylist.sort()
		for k in keylist:
			try:
				forge.targetRegistry[k].help
				forge.cprint(forge.WHITE,0,k)
				print forge.targetRegistry[k].help
			except AttributeError:
				pass

	if forge.printRegistry != 0:
		keylist = forge.targetRegistry.keys()
		keylist.sort()
		for k in keylist:
			print k

	if forge.printDictionary != 0:
		keylist = forge.targetDictionary.keys()
		keylist.sort()
		for k in keylist:
			print k

	forge.cprint(forge.CYAN, 0, "Per Module Timings:")
	for k in forge.buildTimes.keys():
		print "%-30s %5.2f" % (k, forge.buildTimes[k])

	if forge.print_help == 1:
		print forge.UsageHelp()
		for h in forge.help:
			print '-------------------------------------------------------------------------------'
			print h

	forge.cprint(forge.CYAN, 0, forge.get_formatted_elapsedtime())

	forge.modal_quit('forge complete', 0)




###############################################################################
# AS A MODULE
###############################################################################
else:
	import sys
	import os
	import getopt
	import re
	import string
	import imp
	import stat
	import time
	import shutil
	import copy
	import random
	import types
	import popen2
	import threading
	import platform

	help = []
	print_help = 0
	failed = 0
	do_auto = 1
	objlists = {}

	###########################################################################
	def HandleCommandLine(argv):
		"""process the command line"""
		global codegen, recursionDepth, printRegistry, referenceTime, pretty
		global printDictionary, rootFile, printCommon, jobs, blddep, notify
		global rootPath, configModules, print_help, help, do_auto
		try:
				opts, args = getopt.getopt(sys.argv[1:], "xhnmkCctdobr:p:f:F:g:j:", [
				"config=",
				"codegen=",
				"recurse=",
				"auto=",
				"pretty="])
		except getopt.GetoptError:
			cprint(RED,0,'ERROR: invalid command line')
			print UsageHelp()
			sys.exit(2)
		for o, a in opts:
			if o == "-h":
				#print UsageHelp()
				print_help = 1
				#sys.exit(0)
			if o == "-n":
				notify = 1
			if o == "-m":
				print __doc__
				sys.exit(0)
			if o == "-V":
				codegen = 'validate'
			if o == "-d":
				codegen = 'debug'
			if o == "-o":
				codegen = 'optimize'
			if o == "-b":
				blddep = 0
			if o == "-F":
				configModules.append(a)
			if o == "--codegen":
				codegen = a
			if o == "--recurse":
				recursionDepth = int(a)
			if o == "-r":
				recursionDepth = int(a)
			if o == "-f":
				rootFile = a
			if o == "-g":
				rootPath = a
			if o == "-j":
				jobs = int(a)
			if o == "-t":
				printCommon = 1
			if o == "-k":
				printRegistry = 1
			if o == "-c":
				printDictionary = 1
			if o == "-x":
				referenceTime = time.time()
			if o == "--auto":
				if int(a):
					do_auto = 1
				else:
					do_auto = 0
			if o == "--pretty":
				if int(a):
					pretty = 1
				else:
					pretty = 0

			if o == "-p":
				if int(a):
					pretty = 1
				else:
					pretty = 0

		return args


	def UsageHelp():
		help = 'USAGE: [python] ' + sys.argv[0] + ' [options] [targets]\n'
		help +='  \n'
		help +='  OPTIONS:\n'
		help +='  -j <int>			 run up to N jobs in parallel\n'
		help +='  -n				 popup modal notify when forge completes\n'
		help +='  -m				 print manual\n'
		help +='  -d				 debug build\n'
		help +='  -o				 optimized build\n'
		help +='  -V				 validation build (slow debug)\n'
		help +='  --codegen=<string> overrides CODEGEN environment variable\n'
		help +='  --recurse=<int>	 set build recursion depth limit\n'
		help +='  --auto=<0/1>		 do auto configuration processing\n'
		help +='  -b				 suppress target dependence on bld files\n'
		help +='  -r <int>			 set build recursion depth limit\n'
		help +='  -k				 print the target registry\n'
		help +='  -t				 print documented command line targets\n'
		help +='  -c				 print the target dictionary\n'
		help +='  -x				 force a deep rebuild\n'
		help +='  -f <string>		 override root.py as root buildfile name\n'
		help +='  -g <string>		 explicitly set the rootPath\n'
		help +='  --pretty=<0/1>	 use formatted output (default == 1)\n'
		help +='  -p <0/1>			 use formatted output (default == 1)\n'
		help +='  \n'
		help +='  \n'
		help +='  BUILTIN TARGETS:\n'
		help +='  clean				 remove target files\n'
		help +='  here				 module for cwd or \'root\' for root dir\n'
		help +='  root				 root target\n'
		help +='  print				 print stuff for targets\n'
		help +='  \n'
		help +='  \n'
		help +='  ENVIRONMENT VARIABLES:\n'
		help +='  CODEGEN			 set code generation mode\n'
		help +='					 debug, optimize, profile, validate have\n'
		help +='					 default forge settings.  A project may\n'
		help +='					 have more modes supported in its root.py\n'
		help +='  PRETTY			 sets pretty print mode\n'
		help +='  AUTO				 sets auto mode\n'
		help +='  PYTHON			 name of python executable for spawns\n'
		help +='  \n'
		help +='  \n'
		help +='  COMMON EXAMPLES\n'
		help +='  \n'
		help +='	build everything from the root directory:\n'
		help +='	  python forge.py root\n'
		help +='  \n'
		help +='  \n'
		help +='	rebuild everything from the root directory:\n'
		help +='	  python forge.py -x root\n'
		help +='  \n'
		help +='  \n'
		help +='	cleaning libraries and object files just for a module\n'
		help +='	from that module\'s directory:\n'
		help +='	  python forge.py -r2 clean here\n'
		help +='  \n'
		help +='  \n'

		return help
	###########################################################################
	#
	def SetRootPath(path):
		"""reset the root path"""
		global rootPath, rootModule
		rootPath = path
		targetRegistry['root'].AsRootReset()
		targetRegistry['ROOT_BUILD_FILE'] = FileTarget(
			os.path.join(rootPath,rootFile))

		rootModule = Module("")
		targetRegistry['.buildfile'] = targetRegistry['ROOT_BUILD_FILE']

	###########################################################################
	# convienience functions for build rules
	errno = 0
	def quit_callback():
		sys.exit(errno)

	def modal_quit(message, exit_code):
		if not notify:
			return
		import Tkinter
		tk = Tkinter
		root = tk.Tk()

		frame = tk.Frame(root, bg="black")
		frame.pack()

		w = tk.Label(frame, text=message, fg="white", bg="black",
			width=50, height=5)
		w.pack()

		button = tk.Button(frame, text="ok", bg="black", fg="red",
			relief=tk.FLAT, activebackground="black", activeforeground="green",
			command=quit_callback)
		button.pack()

		global errno
		errno = exit_code
		root.mainloop()

	def end_build(name):
		"""color coded failed build"""
		if pretty == 1:
			cprint(RED,1,"failed build: " + name)
		elif pretty != 2:
			print "failed build: " + name

		modal_quit("failed build: " + name, 1)

		sys.exit(1)

	def confirm_path(target):
		(head,tail) = os.path.split(target.name)
		if not os.path.exists(head):
			try:
				os.makedirs(head)
			except OSError:
				print "failed creating " + head
		return tail

	class file_reader(threading.Thread):
		def __init__(self, file):
			threading.Thread.__init__(self)
			self.file = file
		def run(self):
			self.output = self.file.read()

	def execute(command,name):
		if sys.platform == "win32":
			p = os.popen(command,'r')
			output = p.read()
			n = p.close()
			if n != None or pretty == 0:
				print process_output(output, name)
			if n != None:
				end_build(name)
		else:
			p = popen2.Popen3(command, capturestderr=1)
			t_out = file_reader(p.fromchild)
			t_out.start()
			p_err = p.childerr.read()
			t_out.join()
			p_out = t_out.output
			n = os.WEXITSTATUS(p.wait())
			if n != 0 or pretty == 0:
				if p_out != "":
					print process_output(p_out, name)
			if p_err != "" and pretty != 2:
				print process_output(p_err, name)
			if n != 0:
				end_build(name)

	def execute_long(command,arguments,name):
		if sys.platform != "win32":
			return execute(string.join((command,arguments)),name)
		if pretty == 0:
			print "Diverting Arguments to File"
		f = open("tmp.lnk","w")
		f.write(arguments)
		f.close()
		command = string.join((command,"@tmp.lnk"))
		if pretty == 0:
			print command
		return execute(command,name)

	def add_cpps(command, target):
		for d in target.deps:
			match = re.compile(r'(.*)\.(' + c_ext + ')$').match(d.name)
			if match != None:
				cppfile = match.group(0)
				if cppfile != None:
					command = string.join((command,forcecxx + ' ' + cppfile))
		return command

	def add_objs(command, target):
		for d in target.deps:
			match = re.compile(r'(.*)\.(o|obj)$').match(d.name)
			if match != None:
				obj = match.group(0)
				if obj != None:
					command = string.join((command,obj))
		return command

	def add_libs(command, target):
		for d in target.deps:
			match = re.compile(r'.*' + libpre + r'(.*)\.(so|a)$').match(d.name)
			if match != None:
				lib = match.group(1)
				if lib != None:
					command = string.join((command,linkformat%lib))
			match = re.compile(r'.*' + libpre + r'(.*)\.(dylib)$').match(d.name)
			if match != None:
				lib = match.group(1)
				if lib != None:
					command = string.join((command,'-l'+lib))
			match = re.compile(r'(.*)\.(dll|lib|pyd)$').match(d.name)
			if match != None:
				lib = match.group(1)
				if lib != None:
					command = string.join((command,lib+'.lib'))
		return command

	def add_cppflags(command, target):
		"""build cppflags portion of command line"""
		#########################################################
		try :
			target.module.cppflags
			cprint(RED,1,"cppflags deprecated: " + target.name)
			sys.exit(2)
		except AttributeError:
			pass
		#########################################################
		try :
			target.cppflags
			cprint(RED,1,"cppflags deprecated: " + target.name)
			sys.exit(2)
		except AttributeError:
			pass
		#########################################################
		try :
			cppflags
			cprint(RED,1,"cppflags deprecated: " + target.name)
			sys.exit(2)
		except NameError:
			pass
		#########################################################
		#########################################################
		cppDict = copy.deepcopy(cppmap)
		try:
			target.module.cppmap
			for key in target.module.cppmap.keys():
				cppDict[key] = target.module.cppmap[key]
		except AttributeError:
			pass
		try:
			target.cppmap
			for key in target.cppmap.keys():
				cppDict[key] = target.cppmap[key]
		except AttributeError:
			pass
		if target.module != None:
			mod_parts = string.split(target.module.modName, '.')
			if api == 'x86_win32' or api == 'x86_win64':
				cppDict['defmodule'] = '-DMODULE_' + mod_parts[-1]
			cppDict['definemodule'] = '-DMODULE=\\"%s\\"' % target.module.modName
		for key in cppDict.keys():
			#cprint(CYAN,1, 'CPPFLAG: ' + key + ' = ' + cppDict[key])
			command = string.join((command,cppDict[key]))
		#########################################################
		incl_re = re.compile(r'(\-I)(\s*\S+)')
		incls = incl_re.findall(command)
		if len(incls) > 0:
			cprint(RED,1,
				'WARNING: these include paths are specified in cppmap')
		for i in incls:
			cprint(RED,0,"	" + i[1])
		#########################################################
		includeDict = TargetIncludeMap(target)
		keys = includeDict.keys()
		keys.sort()
		for key in keys:
			fullpath = includeDict[key]
#			print 'includemap ' + key + ': ' + fullpath
			if not os.path.isabs(fullpath):
				fullpath = os.path.join(rootPath,fullpath)
				fullpath = os.path.abspath(fullpath)
			command = string.join((command,'-I' + fullpath))
		return command

	def add_linkflags(command, target):
		linkDict = copy.deepcopy(linkmap)
		try:
			target.module.linkmap
			for key in target.module.linkmap.keys():
				linkDict[key] = target.module.linkmap[key]
		except AttributeError:
			pass
		try:
			target.linkmap
			for key in target.linkmap.keys():
				linkDict[key] = target.linkmap[key]
		except AttributeError:
			pass
		for key in linkDict.keys():
			command = string.join((command,linkDict[key]))

		return command

	def TargetIncludeMap(target):
		includeDict = copy.deepcopy(includemap)
		try:
			target.module.includemap
			for key in target.module.includemap.keys():
				includeDict[key] = target.module.includemap[key]
		except AttributeError:
			pass
		try:
			target.includemap
			for key in target.includemap.keys():
				includeDict[key] = target.includemap[key]
		except AttributeError:
			pass
		return includeDict



	def RelativePath(basePath,fullPath):
		remainder = os.path.abspath(fullPath)
		resultPath = None
		while remainder != '/' and remainder != basePath:
			(remainder,tail) = os.path.split(remainder)
			if resultPath:
				resultPath = os.path.join(tail, resultPath)
			else:
				resultPath = tail
		return resultPath


	###########################################################################
	# build rules
	def compile_obj(target):
		"""default object file rule"""

		(modObjPath, tail) = os.path.split(target.name)
		if createpch != '':
			if not hasattr(target, "includemap"):
				target.includemap = {}
			target.includemap["PCH"] = modObjPath

		name = confirm_path(target)
		command = string.join((cxx, nolink, objout+target.name))
		command = add_cppflags(command, target)

		pch_file=''
		pmh_file=''
		post_command=''
		if target.module != None:
			mod_parts = string.split(target.module.modName, '.')
			pch_file = modObjPath + '\\' + mod_parts[-1] + '.' + pch_suffix
			pmh_file = mod_parts[-1] + ".pmh"
		if cxx == 'cl.exe':
			command = string.join((command, ' /Fp' + pch_file))

		for d in target.deps:
			match = re.compile(r'(.*)\.(' + c_ext + ')$').match(d.name)
			if match != None:
				src = match.group(0)
				if src != None:
					matchPMH = re.compile(r'(.*)\.pmh$').match(d.name)
					if matchPMH == None:
						if usepch != '' and os.path.exists(pch_file):
							command = string.join((command,usepch + pmh_file))
						command = string.join((command,forcecxx + ' ' + src))
					elif createpch == '':
						return
					else:
						command = string.join((command,createpch + ' ' + src))
						if cleanpch != '':
							post_command = cleanpch + ' ' + target.name
		if pretty == 1:
			cprint(GREEN,0,'compiling ' + RelativePath(objPath,target.name))
		else:
			print command
		execute(command,name)
		if post_command != '':
			execute(post_command,name)


	def compile_exe(target):
		"""default executable rule"""
		name = confirm_path(target)
		command = string.join((cxx, exeout+target.name))
		command = add_cppflags(command, target)
		command = add_objs(command, target)
		command = add_cpps(command, target)
		command = string.join((command,libincl+libPath))
		command = string.join((command,ld1))
		command = string.join((command,fat1))
		command = add_libs(command, target)
		command = add_linkflags(command, target)
		command = string.join((command,fat2))
		command = string.join((command,ld2))
		if pretty == 1:
			cprint(MAGENTA,0,'compiling ' + name)
		elif pretty == 2:
			pass
		else:
			print command
		execute(command,name)
		if strip != '':
			command = string.join((strip, target.name))
			if pretty == 1:
				cprint(YELLOW,0,'stripping ' + name)
			elif pretty == 2:
				pass
			else:
				print command
			execute(command,name)

	def link_lib(target):
		"""default static library rule"""
#		os.remove(target.name)
		name = confirm_path(target)
		command = ar;
		# this ordering is important '<ar> <flags> <lib> <objs>'
		arguments = string.join((arflags, aroutprefix + target.name))
		arguments = add_objs(arguments, target)
		if pretty == 1:
			cprint(YELLOW,0,'linking ' + name)
		else:
			print command,arguments
		execute_long(command,arguments,name)

	def link_so(target):
		"""default shared library rule"""
		name = confirm_path(target)
		command = '%s -shared -o' % cxx
		command = string.join((command,target.name))
#		command = string.join((command,'-Xlinker --whole-archive'))
		command = add_cppflags(command, target)
		command = add_objs(command, target)
		command = add_cpps(command, target)
		command = string.join((command,'-ldl',libincl+libPath))
		command = add_libs(command, target)
#		command = string.join((command,'-Xlinker --no-whole-archive'))
		command = add_linkflags(command, target)
		if pretty == 1:
			cprint(YELLOW,0,'linking ' + name)
		else:
			print command
		execute(command,name)
		if strip != '':
			command = string.join((strip, target.name))
			if pretty == 1:
				cprint(YELLOW,0,'stripping ' + name)
			elif pretty == 2:
				pass
			else:
				print command
			execute(command,name)

	def link_dylib(target):
		"""default shared library rule"""
		name = confirm_path(target)
		command = '%s -dynamiclib -o ' % cxx
		command = string.join((command,target.name))
		command = add_cppflags(command, target)
		command = add_objs(command, target)
		command = add_cpps(command, target)
		command = string.join((command,'-ldl',libincl+libPath))
		command = add_libs(command, target)
		command = add_linkflags(command, target)
		if pretty == 1:
			cprint(YELLOW,0,'linking ' + name)
		else:
			print command
		execute(command,name)


	def link_dll(target):
		"""default dynamic link library rule"""
		name = confirm_path(target)
		match = re.compile(r'(.*)\.dll').match(target.name)
		command = string.join((linker, linkerflags, dlflags, '/OUT:'+target.name))
		command = string.join((command, '/IMPLIB:'+ match.group(1) + '.lib' ))
		command = add_objs(command, target)
		command = add_cpps(command, target)
		command = add_libs(command, target)
		command = add_linkflags(command, target)
		if pretty == 1:
			cprint(YELLOW,0,'linking ' + name)
		else:
			print command
		execute(command,name)
		if embed != '':
			command = embed % (target.name, target.name)
			if pretty == 1:
				cprint(YELLOW,0,'embedding manifest in ' + name)
			elif pretty == 2:
				pass
			else:
				print command
			execute(command,name)

	def copy_from_srcfile(target):
		(head,tail) = os.path.split(target.name)
		cprint(CYAN,1,'copying ' + tail)
		try:
			src = target.srcfile
		except AttributeError:
			cprint(RED,1,'ERROR: srcfile not specified for ' + target.name)
			return
		try:
			if os.path.exists(src):
				confirm_path(target)
				shutil.copy(src,target.name)
			else:
				cprint(RED,1,'ERROR: could not find ' + src)
		except IOError:
			cprint(RED,1,'ERROR: could not copy from ' + src)

	def noop(target):
		pass

	def build_exe_all(target):
		(head,tail) = os.path.split(target.name)
		cprint(RED,1,"WARNING: " + tail + " falling back to build.exe")
		match = re.compile(r'BUILD:(.*)').match(target.name)
		if match != None:
			dir = match.group(1)
			cwd = os.getcwd()
			os.chdir(dir)
			os.system("build.exe all")
			os.chdir(cwd)

	def build_exe_clean(target):
		(head,tail) = os.path.split(target.name)
		cprint(RED,1,"WARNING: " + tail + " falling back to build.exe")
		match = re.compile(r'BUILD:(.*)').match(target.name)
		if match != None:
			dir = match.group(1)
			cwd = os.getcwd()
			os.chdir(dir)
			os.system("build.exe clean")
			os.chdir(cwd)

	def clean_file(target):
		if isinstance(target,FileTarget):
			target.scanned = 0
			if os.path.exists(target.name):
				(head,tail) = os.path.split(target.name)
				if pretty == 1:
					cprint(MAGENTA,0,'removing ' + tail)
				else:
					print "removing " + target.name
				os.remove(target.name)

	def clean_dll_file(target):
		if isinstance(target,FileTarget):
			target.scanned = 0
			if os.path.exists(target.name):
				(head,tail) = os.path.split(target.name)
				if pretty == 1:
					cprint(MAGENTA,0,'removing ' + tail)
				else:
					print "removing " + target.name
				basename = target.name[:-4]
				os.remove(target.name)
				os.remove(basename + '.ilk')
				os.remove(basename + '.pdb')
				os.remove(basename + '.lib')
				os.remove(basename + '.exp')

	def print_divider(target):
		if pretty == 0:
			print '------------------------------------------------------------'

	def print_pre_target(target):
		global print_depth
		for i in range(0,print_depth):
			sys.stdout.write("|  ")
		print target.name
		print_depth += 1

	def print_post_target(target):
		global print_depth
		print_depth -= 1

	def cctest(file_text, linkmap = None, includemap = None):
		global pretty
		if pretty == 0:
			cprint(YELLOW,1,file_text)

		srctarget = FileTarget(os.path.join(libPath, "forge_auto_test" +".cc"))
		confirm_path(srctarget)

		exetarget = FileTarget(os.path.join(libPath, "forge_auto_test" +".exe"))
		confirm_path(exetarget)


		exetarget.AddDep(srctarget)
		if includemap != None:
			exetarget.includemap = includemap
		if linkmap != None:
			exetarget.linkmap = linkmap
		exetarget.buildNum = -1
		f = open(srctarget.name, 'w')
		f.write(file_text)
		f.close()
		tmp_pretty = pretty
		if pretty == 1:
			pretty = 2
		try:
			exetarget.Build(time.time())
		except:
			pretty = tmp_pretty
			del targetDictionary[srctarget.name]
			del targetDictionary[exetarget.name]
			os.remove(srctarget.name)
			return 'build'

		rc = os.spawnl(os.P_WAIT, exetarget.name, exetarget.name)
		if rc != 0:
			pretty = tmp_pretty
			del targetDictionary[srctarget.name]
			del targetDictionary[exetarget.name]
			os.remove(srctarget.name)
			os.remove(exetarget.name)
			return 'run'


		pretty = tmp_pretty
		del targetDictionary[srctarget.name]
		del targetDictionary[exetarget.name]
		os.remove(srctarget.name)
		os.remove(exetarget.name)
		return None

	def AddSourceDeps(tgt, module, sourcelist):
		for s in sourcelist:
			obj = ObjAndSrcTargets(module,s)
			if obj != None:
				matchPMH = re.compile(r'(.*)\.pmh$').match(s)
				if matchPMH == None:
					obj.threadable = 1
				if tgt != None:
					tgt.AddDep(obj)

	def AddObjDeps(tgt, module, objlist):
		for obj in objlist:
			if obj != None:
				if tgt != None:
					tgt.AddDep(obj)

	def ObjAndSrcTargets(module, src):
		relativePath = RelativePath(rootPath,module.modPath)
		if relativePath:
			modObjPath = os.path.join(objPath, relativePath)
		else:
			modObjPath = objPath
		matchPMH = re.compile(r'(.*)\.pmh$').match(src)
		if matchPMH == None:
			match = re.compile(r'(.*)\.(' + c_ext + ')$').match(src)
			if match == None:
				sourcename = os.path.join(module.modPath,src + cxx_extension)
			else:
				sourcename = os.path.join(module.modPath,match.group(0))
				src = match.group(1)
			objname = os.path.join(modObjPath, src + objsuf)
		elif createpch != '':
			sourcename = os.path.join(module.modPath,src)
			objname = os.path.join(modObjPath, src + '.' + pch_suffix )
		else:
			return None
		if not targetDictionary.has_key(objname):
			obj = FileTarget(objname)
			obj.module = module
			if blddep:
				obj.AddDep(targetRegistry[module.modName + '.buildfile'])
			if not targetDictionary.has_key(sourcename):
				source = FileTarget(sourcename)
				source.module = module
				obj.AddDep(source)
				if blddep:
					source.AddDep(targetRegistry[module.modName + '.buildfile'])
		else:
			obj = targetDictionary[objname]
		return obj


	###########################################################################

	def process_output(output, name):
		for rule in outputContext.GetRules(name):
			output = rule.action(output)
		return output


	###########################################################################
	def setup(modName):
		"""special import function"""
		global moduleDict, failed
		if moduleDict.has_key(modName):
			return
		moduleDict[modName] = 1
		importName = modName + '.bld'
		mod_parts = string.split(modName, '.')
		modPath = rootPath
		for m in mod_parts:
			modPath = os.path.join(modPath,m)
		return mod_setup(modName, modPath)

	def mod_setup(modName, modPath):
		global moduleDict, currentFile, failed, do_auto
		try:
			(file, pathname, description) = imp.find_module('bld', [ modPath ])
		except ImportError:
			cprint(RED,0,"WARNING: " + modName + " bld.py not found")
			buildTarget = BuildExeTarget(modPath)
			moduleDict[modName] = buildTarget
			targetRegistry[modName] = buildTarget
			return

		try:
			pythons_modname = 'forge.' + modName
			parentFile = currentFile
			currentFile = pythons_modname
			py_mod = imp.load_module(pythons_modname,file,pathname,description)
			components = string.split(modName, '.')
			#for comp in components[1:]:
			#	py_mod = getattr(py_mod, comp)
			module = Module(modName)
			module.ok = None # None = ok, otherwise string error message
			if do_auto:
				try:
					if hasattr(py_mod, "auto"):
						module.ok = py_mod.auto(module)
						if module.ok:
							if module.ok[:5] == "abort":
								cprint(RED,1,"ERROR: " + modName +' -- '+module.ok)
								failed = 1
				except Exception, e:
					cprint(RED,1,"ERROR: " + modName + " auto implementation error")
					print e
					module.ok = 'implementation'
			if module.ok == None:
				py_mod.setup(module)
			moduleDict[modName] = module
			targetRegistry[modName] = module
			build_file = FileTarget(os.path.join(modPath,'bld.py'))
			#temporarily disabled to prevent widespread rebuilding bld.py change
			#module.AddDep(build_file)
			#build_file.AddDep(targetRegistry['ROOT_BUILD_FILE'])
			targetRegistry[modName + '.buildfile'] = build_file
			if file != None:
				file.close()

			currentFile = parentFile

			return module

		except Exception, e:
			cprint(RED,1,"WARNING: " + modName + " setup failed")
			print str(e)
			return

	def import_from(modName, modPath):
		re_module_file = re.compile(r'(.+)\.py')
		m = re_module_file.match(modName)
		if m:
			modName = m.group(1)
		(file, pathname, description) = imp.find_module(modName,[ modPath ])

		return imp.load_module(modName, file, pathname, description)

	def deferTo(path_or_depth):
		global path_to_root
		if type(path_or_depth) is types.StringType:
			path_to_root = path
		else:
			for i in range(0,path_or_depth):
				path_to_root = os.path.join(path_to_root,'..')

	def start_timing():
		global t_start
		t_start = time.time()

	def end_timing(target):
		global t_start
		elapsed = time.time() - t_start
		if buildTimes.has_key(target.parentFile):
			buildTimes[target.parentFile] += elapsed
		else:
			buildTimes[target.parentFile] = elapsed

	###########################################################################
	# RULE SUPPORT
	###########################################################################
	class Rule:
		"""Rule base class"""
		def __init__(self, expression, action):
			self.expr = expression
			self.compiledExpr = re.compile(expression,re.M)
			self.action = action
			self.currentMatch = None

	class RuleContext:
		def __init__(self):
			self.ruleList = []
		def AppendRule(self, expression, action):
			rule = Rule(expression, action)
			self.ruleList.append(rule)
		def PrependRule(self, expression, action):
			rule = Rule(expression, action)
			self.ruleList = [rule] + self.ruleList
		def GetRules(self, target):
			ruleSubset = []
			for rule in self.ruleList:
				rule.currentMatch = rule.compiledExpr.match(target)
				if rule.currentMatch:
					ruleSubset.append(rule)
			return ruleSubset


	###########################################################################
	# TARGET SUPPORT
	###########################################################################

	import thread
	import threading
	class BuildThread(threading.Thread):
		def __init__(self,target,ref_time):
			threading.Thread.__init__(self)
			self.target = target
			self.ref_time = ref_time
			self.new_time = ref_time
		def run(self):
			self.new_time = self.target.Build(self.ref_time)

	class Target:
		"""Target base class"""
		def __init__(self,name):
			self.deps = []
			self.backdeps = []
			self.name = name
			if targetDictionary.has_key(self.name):
				raise Exception, 'new target collision: ' + self.name
			targetDictionary[self.name] = self
			self.scanned = 0
			self.buildNum = 0
			self.depContext = None
			self.module = None
			self.threadable = 0
			self.visit = 0
			self.parentFile = currentFile
			self.post_hook = None
		def AppendOverrideRule(self, pattern, method):
			if self.depContext == None:
				self.depContext = RuleContext()
			self.depContext.AppendRule(pattern, method)
		def AddDep(self, dependency):
			self.deps.append(dependency)
			dependency.backdeps.append(self)
		def AddDeps(self, dependencyList):
			for dependency in dependencyList:
				self.AddDep(dependency)
		def RemoveDep(self, dependency):
			self.deps.remove(dependency)
			dependency.backdeps.remove(self)
		def ReplaceDep(self, original, dependency):
			id=self.deps.index(original)
			self.deps.insert(id,dependency)
			self.deps.remove(original)
			original.backdeps.remove(self)
		def Touch(self):
			pass
		def SetTime(self, newTime):
			pass
		def Deps(self):
			return self.deps
		def Exists(self):
			return 0
		def LastBuild(self):
			return 0
		def PrintDeps(self):
			for d in self.deps:
				(head,tail) = os.path.split(d.name)
				print '		' + tail
		def PrintBackDeps(self):
			for d in self.backdeps:
				(head,tail) = os.path.split(d.name)
				print '		' + tail
		def CycleCheck(self):
			cprint(CYAN,1,'cycle check')
			self.Visit()
		def Visit(self):
			if self.visit == 1:
				cprint(RED,1,'WARNING: dependency cycle found')
				cprint(RED,0,'first cycle found is reported.  there may be others')
				cprint(RED,1,'	%s' % (self.name))
				return self
			if self.visit == 2:
				return None

			self.visit = 1

			for d in self.deps:
				v2 = d.Visit()
				if v2 != None:
					cprint(RED,0,'	  <- %s' % (self.name))
					if v2 != self:
						self.visit = 2
						return v2

			self.visit = 2

			return None
			
		def Build(self, reference_time):
			"""\
The basic principal of forge is to ensure that a given target is up
to date.  The purpose of the Build() method of a target is to bring
the target up to date with respect to the given reference time.
An updated reference time is returned.	Typically the returned reference
time is the most recent time the target has been updated or changed.

This method has four main phases:
- Execute matching Pre-Processing Context rules
- Recurse dependency graph, calling Target::Build() on dependencies, getting
  an updated reference time
- If the updated reference time is more recent than the last update time
  for the target, matching Build Context rules are executed
- Execute Post-Processing Context rules
			"""
			(head,tail) = os.path.split(self.name)


			# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
			# cycle protection
			if self.buildNum >= buildNum:
				return self.LastBuild()
			self.buildNum = buildNum

			# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
			# pre context check
			if preContext != None:
				for rule in preContext.GetRules(self.name):
					start_timing()
					rule.action(self)
					end_timing(self)
					

			# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
			# check and build dependencies
			global recursionLevel
			newRefTime = reference_time
			if recursionDepth < 0 or recursionLevel < recursionDepth:
				recursionLevel += 1
				threadables = []
				serials = []
				for d in self.deps:
					if jobs > 0 and d.threadable == 1:
						threadables.append(d)
					else:
						serials.append(d)

				for d in serials:
					newTime = d.Build(reference_time)
					if newTime > newRefTime:
						newRefTime = newTime

				threads = []

				for d in threadables:
					while len(threads) >= jobs:
						threads[0].join()
#						cprint(BLUE,0,'<< ' + threads[0].target.name)
						newTime = threads[0].new_time
						if newTime > newRefTime:
							newRefTime = newTime
						threads.remove(threads[0])
					t = BuildThread(d,reference_time)
					threads.append(t)
#					cprint(BLUE,1,'>> ' + t.target.name)
					t.start()
				for t in threads:
					t.join()
#					cprint(BLUE,0,'<< ' + t.target.name)
					newTime = t.new_time
					if newTime > newRefTime:
						newRefTime = newTime

				recursionLevel -= 1

			if debug_forge:
				print self.name + " lastbuilt:" + time.ctime(
					self.LastBuild()) + " reftime:" + time.ctime(newRefTime)

			# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
			# if necessary build self via actions
			if self.LastBuild() < newRefTime or not self.Exists():
				if self.depContext != None and depContext != None:
					depC = self.depContext
				else:
					depC = depContext
				if depC != None:
					for rule in depC.GetRules(self.name):
						start_timing()
						rule.action(self)
						end_timing(self)
				if self.LastBuild() < newRefTime:
					self.SetTime(newRefTime)

			if not self.Exists():
				self.Touch()

			if self.LastBuild() > newRefTime:
				newRefTime = self.LastBuild()

			# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
			# post context check
			if postContext != None:
				for rule in postContext.GetRules(self.name):
					start_timing()
					rule.action(self)
					end_timing(self)
			return newRefTime


	class FileTarget(Target):
		def __init__(self,name):
			self.name = os.path.abspath(name)
			self.lastBuild = 0
			Target.__init__(self,name)
		def LastBuild(self):
			if self.lastBuild != 0:
				return self.lastBuild
			if os.path.exists(self.name):
				self.lastBuild = os.stat(self.name)[stat.ST_MTIME]
			return self.lastBuild
		def Touch(self):
			self.lastBuild = time.time()
		def SetTime(self,newTime):
			self.lastBuild = newTime
		def Exists(self):
			return os.path.exists(self.name)
		def ShortName(self):
			(head,tail) = os.path.split(self.name)
			return tail

	class CopyTarget(FileTarget):
		def __init__(self,name,src):
			FileTarget.__init__(self,name)
			if isinstance(src,FileTarget):
				self.srcfile = src.name
				srctarget = src
			else:
				self.srcfile = src
				srctarget = LookupTarget(self.srcfile)
				if not srctarget:
					srctarget = FileTarget(self.srcfile)
				srctarget.AppendOverrideRule(".*", noop)
			self.AppendOverrideRule(".*", copy_from_srcfile)
			self.AddDep(srctarget)


	def FindFileTarget(filepath):
		if targetDictionary.has_key(filepath):
			return targetDictionary[filepath]
		else:
			return FileTarget(filepath)

	def FindTarget(name):
		if targetDictionary.has_key(name):
			return targetDictionary[name]
		else:
			return Target(name)

	def LookupTarget(string):
		if targetDictionary.has_key(string):
			return targetDictionary[string]
		return None


	def MatchTarget(arg):
		matches = []
		if arg == 'here':
			if os.getcwd() == rootPath:
				arg = 'root'
			else:
				moduleName = None
				cwd = os.path.abspath(os.getcwd())
				remainder = cwd
				resultPath = None
				while remainder != rootPath:
					(remainder,tail) = os.path.split(remainder)
					if moduleName:
						moduleName = tail + '.' + moduleName
					else:
						moduleName = tail
					if remainder == '/':
						cprint(YELLOW,1,"Can't find " + cwd + " as 'here'")
						cprint(YELLOW,1,"\tin " + rootPath + " -> using 'root'")
						moduleName = 'root'
						break
				arg = moduleName

		if targetRegistry.has_key(arg):
			matches.append(targetRegistry[arg])
			return targetRegistry[arg]
		x = os.path.join(os.getcwd(),arg)
		if targetRegistry.has_key(x):
			return targetRegistry[x]
		return None

	###########################################################################
	# DEPENDENCY SUPPORT
	###########################################################################


	def FindInPath(target, filename):
		# first check file's local directory
		(head,tail) = os.path.split(target.name)
		filepath = os.path.join(head, filename)
		if os.path.isfile( filepath ):
			return filepath

		# then check the specified include directories
		include_map = TargetIncludeMap(target)
		for suppress_key in includeScanSuppress:
			del include_map[suppress_key]
		for key in include_map.keys():
			fullpath = include_map[key]
			if not os.path.isabs(fullpath):
				fullpath = os.path.join(rootPath,fullpath)
				fullpath = os.path.abspath(fullpath)
			filepath = os.path.join(fullpath, filename)
			if os.path.isfile( filepath ):
				return filepath
		return None

	def ScanIncludes(fileTarget):
		#print 'scan ' + fileTarget.name
		if fileTarget.scanned == 1:
			return
		fileTarget.scanned = 1
		isPCH = re.compile(r'.*\..?ch$').match(fileTarget.name)
		if isPCH != None:
			return
		if not os.path.isfile( fileTarget.name ):
			return
		f = open(fileTarget.name)
		content = f.read()
		f.close()
		includes = include_re.findall(content)
		for i in includes:
			#print i[1]
			filepath = FindInPath(fileTarget,i[1])
			if filepath != None:
				includeTarget = FindFileTarget(filepath)
				#print 'bind ' + fileTarget.name + ' -> ' + includeTarget.name
				fileTarget.AddDep(includeTarget)
				ScanIncludes(includeTarget)
			else:
				#cprint(RED,0,"		 " + i[1] + " " + fileTarget.ShortName())
				pass

	def AddScanInclude(include_path):
		if not os.path.isabs(include_path):
			include_path = os.path.join(rootPath, include_path)
			include_path = os.path.abspath(include_path)
		scanIncludePath.append(include_path)
		return include_path

	def dep_lookup(consumer, provider):
		if isinstance(consumer,Target):
			c = consumer
		else:
			c = targetRegistry[consumer]
		if isinstance(provider,Target):
			p = provider
		else:
			p = targetRegistry[provider]
		c.AddDep(p)

	def deps(consumers, providers):
		dep(consumers, providers)

	def dep(consumers, providers):
		if isinstance(consumers, types.ListType):
			consumer_list = consumers
		else:
			consumer_list = [ consumers ]

		if isinstance(providers, types.ListType):
			provider_list = providers
		else:
			provider_list = [ providers ]

		for consumer in consumer_list:
			for provider in provider_list:
				dep_lookup(consumer, provider)

	class PlaceholderTarget(Target):
		def __init__(self, name):
			Target.__init__(self, name)
		def Build(self, reference_time):
			cprint(RED,1,"ERROR: unresolved placeholder target: " + self.name)
			sys.exit(1)

	class TargetRegistry:
		def __init__(self):
			self.map = {}
		def __setitem__(self, key, value):
			if self.map.has_key(key):
				oldvalue = self.map[key]
				if isinstance(oldvalue, PlaceholderTarget):
					for d in oldvalue.deps:
						value.AddDep(d)
					while len(oldvalue.backdeps) > 0:
						d = oldvalue.backdeps[0]
						d.ReplaceDep(oldvalue,value)
					oldvalue.PrintBackDeps()
				else:
					cprint(RED,1,"ERROR: attempt to redefine target: " + key)
					sys.exit(1)
			self.map[key] = value
		def __getitem__(self, key):
			if self.map.has_key(key):
				return self.map[key]
			placeholder = PlaceholderTarget(key)
			self.map[key] = placeholder
			return placeholder
		def has_key(self, key):
			return self.map.has_key(key)
		def keys(self):
			return self.map.keys()

	###########################################################################
	class Module(FileTarget):
		"""Module target class. A module is a directory with its own
		bld.py file."""
		def __init__(self, modName):
			self.modName = modName
			dirstruct = string.split(modName, '.')
			self.modPath = rootPath
			for dir in dirstruct:
				self.modPath = os.path.join(self.modPath,dir)
			self.modPath = os.path.abspath(self.modPath)
			FileTarget.__init__(self, self.modPath)
			targetRegistry['root'].AddDep(self)
			self.hasDLL = 0
			self.hasLib = 0
			self.hasExe = 0
		def Module(self, subName):
			n = setup(self.modName + '.' + subName)
			if n != None:
				self.AddDep(n)
			return n
		def SrcTargetName(self,SrcnameWithExtension):
			return os.path.join(self.modPath,SrcnameWithExtension)
		def ObjTargetName(self,ObjNameWithoutExtension):
			modObjPath = os.path.join(objPath, RelativePath(rootPath,self.modPath))
			return os.path.join(modObjPath, ObjNameWithoutExtension + objsuf)
		def Exe(self, name, srclist = None):
			self.hasExe = 1
			if srclist == None:
				srclist = [ name ]
			exeTarget = FileTarget(os.path.join(libPath,name + ".exe"))
			exeTarget.module = self
			exeTarget.threadable = 1
			targetRegistry[name + 'Exe'] = exeTarget
			AddSourceDeps(exeTarget,self,srclist)
			self.AddDep(exeTarget)
			if blddep:
				exeTarget.AddDep(targetRegistry[self.modName + '.buildfile'])
			return exeTarget
		def Lib(self, name, srclist, registeredName = None):
			self.hasLib = 1
			libTarget = FileTarget(os.path.join(libPath,libpre + name + libsuf))
			libTarget.module = self
			#global targetRegistry
			if registeredName != None:
				targetRegistry[registeredName] = libTarget
			else:
				targetRegistry[name + 'Lib'] = libTarget
			AddSourceDeps(libTarget,self,srclist)
			self.AddDep(libTarget)
			if blddep:
				libTarget.AddDep(targetRegistry[self.modName + '.buildfile'])
			return libTarget
		def SrcToObj(self, srcList):
			objlist = []
			for s in srcList:
				t = ObjAndSrcTargets(self, s)
				if t != None:
					objlist.append(t)
			return objlist
		def LibFromObjs(self, name, objlist, registeredName = None):
			self.hasLib = 1
			libTarget = FileTarget(os.path.join(libPath,libpre + name + libsuf))
			libTarget.module = self
			if registeredName != None:
				targetRegistry[registeredName] = libTarget
			else:
				targetRegistry[name + 'Lib'] = libTarget
			AddObjDeps(libTarget,self,objlist)
			self.AddDep(libTarget)
			if blddep:
				libTarget.AddDep(targetRegistry[self.modName + '.buildfile'])
			return libTarget
		def DLL(self, name, srclist, registeredName = None):
			self.hasDLL = 1
			libTarget = FileTarget(os.path.join(libPath,libpre + name + libdl))
			libTarget.module = self
			if registeredName != None:
				targetRegistry[registeredName] = libTarget
			else:
				targetRegistry[name + 'Lib'] = libTarget
			AddSourceDeps(libTarget,self,srclist)
			self.AddDep(libTarget)
			if blddep:
				libTarget.AddDep(targetRegistry[self.modName + '.buildfile'])
			return libTarget
		def Build(self, reference_time):
			FileTarget.Build(self,reference_time)

	class RootModule(Module):
		def __init__(self):
			self.modName = 'root'
			self.modPath = rootPath
			FileTarget.__init__(self, self.modPath)
		def AsRootReset(self):
			if targetDictionary.has_key(self.name):
				del targetDictionary[self.name]
			self.modPath = rootPath
			self.name = self.modPath

	class CleanTarget(Target):
		def __init__(self,str):
			Target.__init__(self,str)
			self.help = """\
remove target files
example to clean all target files:
  forge.py clean root
"""
		def Build(self, reference_time):
			global preContext
			global depContext
			if preContext:
				cleanContext = preContext
			else:
				preContext = defaultCleanContext
			depContext = None

	class BuildTarget(Target):
		def Build(self, reference_time):
			global preContext
			global depContext
			depContext = defaultContext


	class PrintTarget(Target):
		def __init__(self,str):
			Target.__init__(self,str)
			self.help = """\
print stuff for targets
this is useful for debugging, although somewhat verbose
"""
		def Build(self, reference_time):
			global preContext
			global depContext
			global postContext
			if preContext:
				printPreContext = preContext
			else:
				printPreContext = RuleContext()
				preContext = printPreContext
			if postContext:
				printPostContext = preContext
			else:
				printPostContext = RuleContext()
				postContext = printPostContext

			printPreContext.AppendRule(".*", print_pre_target)
			printPostContext.AppendRule(".*", print_post_target)

	class BuildExeTarget(Target):
		def __init__(self, name):
			self.lastBuild = 0
			Target.__init__(self, "BUILD:" + name)
			targetRegistry['root'].AddDep(self)
			(head,tail) = os.path.split(name)
			targetRegistry["build.exe:" + tail] = self
		def LastBuild(self):
			return self.lastBuild
		def Touch(self):
			self.lastBuild = time.time()
		def SetTime(self,newTime):
			self.lastBuild = newTime

	def ScanAllTargets():
		cprint(CYAN,0,'begin scanning dependencies')
		for key in targetDictionary.keys():
			if re.compile(r'(.*)\.(' + c_ext + ')').match(key):
				ScanIncludes(targetDictionary[key])
		depcnt = 0
		for t in targetDictionary.values():
			depcnt += len(t.deps)
		cprint(CYAN,0,'%d targets' % (len(targetDictionary)))
		cprint(CYAN,0,'%d dependencies' % (depcnt))
		cprint(CYAN,0,'end scanning dependencies')


	###########################################################################
	###########################################################################
	# forge module load time code
	###########################################################################
	version_match = re.compile(r'(\d*\.\d*).*').match(sys.version)
	if not version_match:
		cprint(forge.RED,1,'ERROR: cannot determine python version')
	else:
		python_version = float(version_match.group(1))

	if sys.platform == "win32":
		wconioColor = 0
		try:
			import WConio
			wconioColor = 1
		except ImportError:
			print "\n\n\n\n\n"
			print "forge.py: install the package WConio to get color output"
			print "\n\n\n\n\n"

	###########################################################################
	# defaults
	codegen = 'default'
	compiler = 'echo'
	pretty = 1
	blddep = 1

	###########################################################################
	# initializations

	debug_forge = None
	scanIncludePath = []
	targetDictionary = {}
	preContext = None
	depContext = None
	postContext = None
	buildNum = 0
	dlflags = ""
	moduleDict = {}
	cppmap = {}
	linkmap = {}
	includemap = {}
	includeScanSuppress = []
	rootPath = os.getcwd()
	libPath = rootPath
	objPath = rootPath
	binPath = rootPath
	print_depth = 0
	# seems better to not chase system includes
	#include_re = re.compile(r'^[ \t]*#[ \t]*include[ \t]*(<|")([\w./\\]+)(>|")', re.M)
	include_re = re.compile(r'^[ \t]*#[ \t]*include[ \t]*(")([\w./\\]+)(")', re.M)
	recursionDepth = -1
	recursionLevel = 0
	printRegistry = 0
	notify = 0
	printCommon = 0
	printDictionary = 0
	cxx_extension = '.cc' # when creating transient targets
	c_ext = 'c|cc|cpp|C|cxx|pmh'
	referenceTime = 0
	path_to_root = ""

	###########################################################################
	# color support
	if sys.platform == "win32" and wconioColor == 1:
		RED		= WConio.RED
		GREEN	= WConio.GREEN
		YELLOW	= WConio.YELLOW
		BLUE	= WConio.BLUE
		MAGENTA	= WConio.MAGENTA
		CYAN	= WConio.CYAN
		WHITE	= WConio.WHITE
	else:
		RED		= 31
		GREEN	= 32
		YELLOW	= 33
		BLUE	= 34
		MAGENTA	= 35
		CYAN	= 36
		WHITE	= 37
		NORMAL	= 0
		BOLD	= 1

	def ansi(c):
		return "[%dm" % c

	def color_on(i, j):
		if(sys.platform == "win32"):
			if wconioColor == 1:
				WConio.textcolor(j)
				if i == 0:
					WConio.lowvideo()
				else:
					WConio.highvideo()
		else:
			sys.stdout.write( r"[%dm[%dm" % (i,j) )

	def color_off():
		if(sys.platform == "win32"):
			if wconioColor == 1:
				WConio.normvideo()
		else:
			sys.stdout.write( "[0m" )
			sys.stdout.flush()

	print_lock = threading.Lock()
	def cprint(color, bold, string):
		print_lock.acquire()
		color_on(bold,color)
		sys.stdout.write(string + '\n')
		color_off()
		print_lock.release()

	def cprint_no_newline(color, bold, string):
		print_lock.acquire()
		color_on(bold,color)
		sys.stdout.write(string)
		color_off()
		print_lock.release()

	def initialize():
		global codegen,pretty,regex,filesystem
		global PythonExe,args,api,cxx,objout,exeout,cppmap
		global libincl,nolink,strip,embed,ar,arflags,aroutprefix,objsuf,exesuf
		global linkformat,libpre,libsuf,libdl,pymodsuf,ld1,ld2
		global forcecxx,createpch,cleanpch,usepch,pch_suffix
		global compiler_version,cpp_0x,fat1,fat2,winlibs
		global gllibs,targetRegistry,defaultContext,depContext,rootFile,jobs
		global defaultCleanContext,cleanTarget,bldTarget,rootPath,dlflags
		global linker, linkerflags, ccache, distcc, blddep
		global outputContext, configModules, currentFile, buildTimes, do_auto
		global fe_config, module_list, lab_list
		#######################################################################
		# environment variables
		if os.environ.has_key("CODEGEN"):
			codegen = os.environ["CODEGEN"]
		if os.environ.has_key("PRETTY"):
			pretty = int(os.environ["PRETTY"])
		if os.environ.has_key("AUTO"):
			do_auto = int(os.environ["AUTO"])
		if os.environ.has_key("PYTHON"):
			PythonExe = os.environ["PYTHON"]
		else:
			PythonExe = 'python'

		regex = 1
		if os.environ.has_key("FE_BOOST_REGEX"):
			if int(os.getenv('FE_BOOST_REGEX')) == 1:
				cprint(CYAN,0,'boost regex on')
				cppmap['regex'] = '-DFE_BOOST_REGEX=1'
			else:
				cprint(CYAN,0,'boost regex off')
				cppmap['regex'] = '-DFE_BOOST_REGEX=0'
				regex = 0

		filesystem = 1
		if os.environ.has_key("FE_BOOST_FILESYSTEM"):
			if int(os.getenv('FE_BOOST_FILESYSTEM')) == 1:
				cprint(CYAN,0,'boost filesystem on')
				cppmap['filesystem'] = '-DFE_BOOST_FILESYSTEM=1'
			else:
				cprint(CYAN,0,'boost filesystem off')
				cppmap['filesystem'] = '-DFE_BOOST_FILESYSTEM=0'
				filesystem = 0

		if os.environ.has_key("FE_TSS"):
			if int(os.getenv('FE_TSS')) == 1:
				cprint(CYAN,0,'thread safe storage on')
				cppmap['tss'] = '-DFE_TSS=1'
			else:
				cprint(CYAN,0,'thread safe storage off')
				cppmap['tss'] = '-DFE_TSS=0'

		#######################################################################
		rootFile = 'root.py'
		jobs = 0
		configModules = []
		currentFile = 'forge'
		buildTimes = {}
		args = HandleCommandLine(sys.argv)
		strip = ''
		embed = ''
		forcecxx = ''
		createpch = ''
		cleanpch = ''
		usepch = ''
		pch_suffix = ''
		fe_config = {}
		module_list = []
		lab_list = []

		#######################################################################
		# main platform dependent setup
		# LINUX G++ #######################################

		determine_api()

		if api == "x86_linux" or api == "x86_64_linux":
			cxx = ''
			if os.path.isfile( '/usr/bin/ccache' ):
				cxx += 'ccache '
				ccache=1
			if os.path.isfile( '/usr/bin/distcc' ) and os.environ.has_key("DISTCC_HOSTS") and os.environ["DISTCC_HOSTS"] != "":
				if ccache:
					cxx = 'CCACHE_PREFIX="distcc" ' + cxx
				else:
					cxx += 'distcc '
				distcc=1
			if os.environ.has_key("FE_CC"):
				cxx += os.environ["FE_CC"]
			else:
				cxx += 'g++'
			# fails on athlon (duh):
			#cxx += ' -fPIC -march=pentium3 '
			cxx += ' -fPIC'
			objout = '-o'
			exeout = '-o'

			# TODO consider -pedantic
			if codegen == 'validate':
				cppmap['codegen'] = '-D_DEBUG -g3 -Wall -Wextra'
			if codegen == 'debug':
				cppmap['codegen'] = '-D_DEBUG -g2 -Wall -Wextra'
			if codegen == 'fast':
				cppmap['codegen'] = '-D_DEBUG -Wall -O0'
			if codegen == 'profile':
				cppmap['codegen'] = '-O3 -Wall'
				# cppmap['codegen'] += ' -g'

				# gprof profiling
				#cppmap['codegen'] += ' -g -pg'
			if codegen == 'optimize':
				#cppmap['codegen'] = '-march=pentium3 -msse -mmmx -mfpmath=sse'
				#cppmap['codegen'] = '-g -O3 -malign-double -m64'
				# above had " error: -malign-double makes no sense in the 64bit mode"

				cppmap['codegen'] = '-O3 -Wall'
				# cppmap['codegen'] += ' -g'

				#strip = 'strip --strip-unneeded'

			if api == "x86_64_linux":
				cppmap['codegen'] += ' -m64'

			# whole lotta these; block now, but we should fix eventually
			cppmap['codegen'] += ' -Wno-unused-parameter'

			# questionable approximations
#			cppmap['codegen'] += ' -ffast-math'

			libincl = '-L'
			nolink = '-c'
			ar = 'ar'
			arflags = '-ur'
			aroutprefix = ''
			objsuf = '.o'
			exesuf = '.exe'
			linkformat = '-l%s'
			libpre = 'lib'
			libsuf = '.a'
			libdl = '.so'
			pymodsuf = '.so'
			ld1 = '-ldl'
			ld2 = ''
#			forcecxx = '-x c++'		# distcc doesn't like
			fat1 = '-rdynamic -Xlinker --export-dynamic -Xlinker --whole-archive'
			fat2 = '-Xlinker --no-whole-archive'

			winlibs = ''
			gllibs = ''

		# PPC linux ###########################################
		elif api == "ppc_linux":
			cxx = ''
			if os.path.isfile( '/usr/bin/ccache' ):
				cxx += 'ccache '
				ccache=1
			if os.path.isfile( '/usr/bin/distcc' ) and os.environ.has_key("DISTCC_HOSTS") and os.environ["DISTCC_HOSTS"] != "":
				cxx += 'distcc '
				distcc=1
			if os.environ.has_key("FE_CC"):
				cxx += os.environ["FE_CC"]
			else:
				cxx += 'g++'
			# fails on athlon (duh):
			#cxx += ' -fPIC -march=pentium3 '
			cxx += ' -fPIC'
			objout = '-o'
			exeout = '-o'

			# TODO consider -pedantic
			if codegen == 'validate':
				cppmap['codegen'] = '-D_DEBUG -g3 -Wall -Wextra'
			if codegen == 'debug':
				cppmap['codegen'] = '-D_DEBUG -g2 -Wall -Wextra'
			if codegen == 'fast':
				cppmap['codegen'] = '-D_DEBUG -Wall -O0'
			if codegen == 'profile':
				cppmap['codegen'] = '-O3 -Wall'
				cppmap['codegen'] += ' -g'
				# gprof profiling
				#cppmap['codegen'] = ' -pg'
			if codegen == 'optimize':
				#cppmap['codegen'] = '-march=pentium3 -msse -mmmx -mfpmath=sse'
				cppmap['codegen'] = '-O3'
				strip = 'strip --strip-unneeded'

			# whole lotta these; block now, but we should fix eventually
			cppmap['codegen'] += ' -Wno-unused-parameter'

			# questionable approximations
#			cppmap['codegen'] += ' -ffast-math'

			libincl = '-L'
			nolink = '-c'
			ar = 'ar'
			arflags = '-ur'
			aroutprefix = ''
			objsuf = '.o'
			exesuf = '.exe'
			linkformat = '-l%s'
			libpre = 'lib'
			libsuf = '.a'
			libdl = '.so'
			pymodsuf = '.so'
			ld1 = '-ldl'
			ld2 = ''
#			forcecxx = '-x c++'		# distcc doesn't like
			fat1 = '-rdynamic -Xlinker --export-dynamic -Xlinker --whole-archive'
			fat2 = '-Xlinker --no-whole-archive'

			winlibs = ''
			gllibs = ''

		# CYGWIN ###########################################
		elif api == "x86_cygwin":
			cxx = 'g++'
			objout = '-o'
			exeout = '-o'
			if codegen == 'debug' or codegen == 'validate':
				cppmap['codegen'] = '-g3 -Wall'
			if codegen == 'optimize':
				#cppmap['codegen'] = '-O3 -march=pentium3 -msse -mmmx -mfpmath=sse'
				cppmap['codegen'] = '-O3'

			libincl = '-L'
			nolink = '-c'
			ar = 'ar'
			arflags = '-ur'
			aroutprefix = ''
			objsuf = '.o'
			exesuf = '.exe'
			linkformat = '-l%s'
			libpre = 'lib'
			libsuf = '.a'
			libdl = '.so'
			pymodsuf = '.so'
			ld1 = ''
			ld2 = ''
			fat1 = '-rdynamic -Xlinker --export-dynamic -Xlinker --whole-archive'
			fat2 = '-Xlinker --no-whole-archive'

			winlibs = ''
			gllibs = ''

		# WIN32 VC ########################################
		elif api == "x86_win32" or api == "x86_win64":
			cxx = 'cl.exe'
			objout = '/Fo'
			exeout = '/Fe'
			dlflags += ' /DLL'
			if codegen == 'debug' or codegen == 'validate':
				# /MDd causes initialization failure dialogs (VS express)
				# cppmap['codegen'] = '/D _DEBUG /MTd /MDd /Od /Gm- /Zi /GZ'
				#cppmap['codegen'] = '/D _DEBUG /MTd /MD /Od /Gm- /Zi /GZ'
				cppmap['codegen'] = '/D _DEBUG /MD /Zi /EHsc'
				dlflags += ' /DEBUG'
			elif codegen == 'optimize':
				#cppmap['codegen'] = '/D NDEBUG /MT /MD /O2 /EHsc'
				# crap, this one is the only one for now w/ TBB:
				#cppmap['codegen'] = '/D _DEBUG /MD /O2 /EHsc'
				#cppmap['codegen'] = '/MT /GS- /GF /fp:precise /O2 /EHsc /Ob1 /Oi /Ot /Oy /GT'

        # 34s
				#cppmap['codegen'] = '/D _DEBUG /MD /GS- /GF /fp:precise /O2 /EHsc /Ob1 /Oi /Ot /Oy /GT'

        # 16s
				#cppmap['codegen'] = '/MD /GS- /GF /O2 /EHsc- /Ob1 /Oi /Ot /Oy'

				cppmap['codegen'] = '/MD /O2 /Zi /fp:fast'
				#cppmap['codegen'] = '/MD /O2'
			else:
				cppmap['codegen'] = '/MT'
			cppmap['codegen'] += ' /GR /Zc:ForScope'
			cprint(RED, 1, "======================================================")
			cprint(RED, 1, "======================================================")
			cprint(RED, 0, "WARNING: FE_MEM_RAWTEST is set")
			cprint(RED, 1, "======================================================")
			cprint(RED, 1, "======================================================")
			cppmap['codegen'] += ' /D FE_MEM_RAWTEST'
			libincl = '/LIBPATH:'
			nolink = '/c'
			linker = 'link.exe'
			linkerflags = '/NOLOGO /DEBUG'
			#ar = 'lib.exe'
			ar = 'link.exe /LIB'
			arflags = '/NOLOGO'
			aroutprefix = '/OUT:'
			objsuf = '.obj'
			exesuf = '.exe'
			linkformat = '%s.lib'
			libpre = ''
			libsuf = '.lib'
			libdl = '.dll'
			pymodsuf = '.pyd'
			ld1 = '/link'
			ld2 = ''
			forcecxx = '-Tp'

			# TODO fix
#			createpch = '/Yc -Tp'
#			cleanpch = 'del'
#			usepch = '/Yu'

			pch_suffix = 'pch'
			fat1 = ''
			fat2 = ''
			embed = 'mt.exe -manifest %s.manifest -outputresource:%s;2'

			winlibs = ''
			gllibs = ''

			#cppmap['win32'] ='/nologo /EHsc /GX'
			cppmap['win32'] ='/nologo /EHsc-'
			if api == "x86_win64":
				cppmap['win64'] = ' /D "WIN64"'
			else:
				cppmap['win32'] += ' /D "WIN32"'
			cppmap['win32'] += ' /D "WIN32_LEAN_AND_MEAN"'
			cppmap['win32'] += ' /D "_MBCS" /FD'
			cppmap['win32'] += ' /D "_LIB"'
			cppmap['win32_warning'] = '/W3'
			cppmap['win64_warning'] = '/Wp64'

			sse_version = 0.0
			if os.environ.has_key("FE_USE_SSE"):
				sse_version = eval(os.getenv('FE_USE_SSE'))
				fe_config['FE_USE_SSE'] = int(sse_version)

			if sse_version >= 1.0:
				cprint(CYAN,0,'Using SSE' + str(sse_version))
				if sse_version >= 2.0:
					cppmap['sse'] += ' /arch:SSE2'
				else:
					cppmap['sse'] += ' /arch:SSE'
			cppmap['sse'] = '/arch:SSE2'
		# MAC OSX #########################################
		elif api == "ppc_osx" or api == "x86_osx":
			cxx = ''
			if os.environ.has_key("FE_CC"):
				cxx += os.environ["FE_CC"]
			else:
				cxx += 'g++'
			cxx += ' -fPIC -dynamic'
			objout = '-o '
			exeout = '-o '
			if codegen == 'validate':
				cppmap['codegen'] = '-D_DEBUG -g3 -Wall'
			if codegen == 'debug':
				cppmap['codegen'] = '-D_DEBUG -g2 -Wall'
			if codegen == 'fast':
				cppmap['codegen'] = '-D_DEBUG -Wall -O0'
			if codegen == 'profile':
				cppmap['codegen'] = '-O3 -Wall'
				cppmap['codegen'] += ' -g'
			if codegen == 'optimize':
				cppmap['codegen'] = '-O3'
				strip = 'strip -S -x'


			libincl = '-L'
			nolink = '-c'
			ar = 'ar'
			arflags = '-ur'
			aroutprefix = ''
			objsuf = '.o'
			exesuf = '.exe'
			linkformat = '-l%s'
			libpre = 'lib'
			libsuf = '.a'
			libdl = '.dylib'
			pymodsuf = '.so'
			ld1 = '-ldl'
			ld2 = ''
			fat1 = '-bind_at_load'
			fat2 = ''

			winlibs = ''
			gllibs = ''
		else:
			print "unsupported platform"
			sys.exit(1)

		if api == "x86_linux" or api == "x86_64_linux" or api == "ppc_osx" or api == "x86_osx" or api == "ppc_linux":
			command = string.replace(cxx,"ccache","") + " --version"
			p = os.popen(command,'r')
			output = p.read()
			n = p.close()
			lines = string.split(output,'\n')
			words = string.split(lines[0])
			compiler_version = 0.0
			cpp_0x = 0
			cppmap['sse'] = ''
			if len(words)>2:
				digits = string.split(words[len(words)-1],'.')
				compiler_version = float(digits[0] + '.' + digits[1])
			cprint(CYAN,0,"GCC version = " + str(compiler_version))
			if compiler_version >= 4.3:
# boost 1.35 not ready (debug seems to work; optimize refuses to compile)
#				cprint(CYAN,0,"C++0x: supported")
#				cppmap['codegen'] += ' -std=c++0x'
#				cpp_0x = 1
# so instead
				cprint(CYAN,0,"C++0x: boost not ready")
				cppmap['codegen'] += ' -Wno-deprecated'
			else:
				cprint(CYAN,0,"C++0x: not supported by compiler")
			if compiler_version >= 4.0:
				use_visibility=os.getenv('FE_USE_VISIBILITY')
				if use_visibility == '1':
					cprint(CYAN,0,'Using Gnu Class Visibility')
					cppmap['codegen'] += ' -DGCC_HASCLASSVISIBILITY'
					cppmap['codegen'] += ' -fvisibility-inlines-hidden'
					cppmap['codegen'] += ' -fvisibility=hidden'
			if compiler_version >= 3.4:
				use_pch=os.getenv('FE_USE_PCH')
				if use_pch == '1':
					cprint(CYAN,0,'Using Gnu PCH')
					cppmap['codegen'] += ' -Winvalid-pch'
					createpch = '-x c++-header'
					pch_suffix = 'gch'
			if compiler_version >= 3.3:
				sse_version = 0.0
				if os.environ.has_key("FE_USE_SSE"):
					sse_version = eval(os.getenv('FE_USE_SSE'))
					fe_config['FE_USE_SSE'] = int(sse_version)

				# arch settings are by what we use,
				# not by properly determining the target system
				if sse_version >= 1.0:
					cprint(CYAN,0,'Using Gnu SSE' + str(sse_version))
					if sse_version < 2.0:
						cppmap['sse'] += ' -msse -mfpmath=sse'
						# for prefetch
#						cppmap['sse'] += ' -mtune=pentium3'
					else:
#						cppmap['sse'] += ' -msse -msse2 -mmmx -mfpmath=sse'
						cppmap['sse'] += ' -msse' + str(sse_version)
						cppmap['sse'] += ' -pipe'
						if sse_version < 3.0 or compiler_version < 4.3:
							cppmap['sse'] += ' -march=k8'
						else:
							cppmap['sse'] += ' -march=core2'

		targetRegistry = TargetRegistry()

		#######################################################################
		# Default Context
		#######################################################################
		defaultContext = RuleContext()
		defaultContext.AppendRule(r"(.*)\.(obj|o|pch|gch)", compile_obj)
#		defaultContext.PrependRule(r"(.*)", print_divider)
		defaultContext.AppendRule(r"(.*)\.exe", compile_exe)
		defaultContext.AppendRule(r"(.*)\.(lib|a)", link_lib)
		defaultContext.AppendRule(r"(.*)\.so", link_so)
		defaultContext.AppendRule(r"(.*)\.dylib", link_dylib)
		defaultContext.AppendRule(r"(.*)\.dll", link_dll)
		defaultContext.AppendRule(r"BUILD:(.*)", build_exe_all)
		depContext = defaultContext

		defaultCleanContext = RuleContext()
		defaultCleanContext.AppendRule(r"(.*)\.(obj|o|exe|lib|a|so|pch|gch|dylib)", clean_file)
		defaultCleanContext.AppendRule(r"(.*)\.(dll)", clean_dll_file)
		defaultCleanContext.AppendRule(r"BUILD:(.*)", build_exe_clean)

		outputContext = RuleContext()

		cleanTarget = CleanTarget("clean")
		targetRegistry['clean'] = cleanTarget
		bldTarget = BuildTarget("build")
		targetRegistry['build'] = bldTarget
		printTarget = PrintTarget("print")
		targetRegistry['print'] = printTarget
		if not rootPath:
			rootPath = os.getcwd()
		rootTarget = RootModule()
		targetRegistry['root'] = rootTarget
		rootTarget.help = """\
root target
"""


	def determine_api():
		global api

		hw = string.lower(platform.machine())
		os_name = string.lower(platform.system())

		# normalize hardware identifiers
		if hw == 'i686' or hw == 'i386' or hw == 'amd64' or hw == '':
			hw = 'x86'
		if hw == 'power macintosh':
			hw = 'ppc'

		if os_name == 'windows':
#			(release, version, csd, ptype) = platform.win32_ver()
			env_path = os.getenv('Path')
			first = env_path.find("BIN\\amd64")
			if first >= 0:
				os_name = 'win64'
			else:
				os_name = 'win32'
		if os_name == 'darwin':
			os_name = 'osx'

		api = hw + '_' + os_name

		cprint(CYAN,0,"API = " + api)

	def reset_buildtime():

		global start_buildtime
		start_buildtime = time.time()

	def get_elapsed_buildtime():

		global start_buildtime,end_buildtime

		end_buildtime = time.time()
		return (end_buildtime - start_buildtime)


	def get_formatted_elapsedtime():

		import math
		duration = get_elapsed_buildtime()
		(temp,hours)   = math.modf(duration / 3600.0)
		(temp,minutes) = math.modf(temp*60.0)
		(temp,seconds) = math.modf(temp*60.0)

		return "Total Elapsed build time: %02d:%02d:%02d   (total seconds: %.2f)" % (hours,minutes,seconds,duration)


