/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#define FE_SDO_DEBUG	FALSE

#include "operator/operator.pmh"

namespace fe
{

void SubdivideOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	catalog<String>("partitionAttr")="hull_primitive";
	catalog<String>("partitionAttr","label")="Partition Attr";
	catalog<String>("partitionAttr","hint")=
			"Primitive string attribute created on output"
			" distinguishing the original primitives.";

	catalog<bool>("recomputeNormals")=true;
	catalog<String>("recomputeNormals","label")="Recompute Normals";

	catalog<I32>("depth")=1;
	catalog<I32>("depth","high")=8;
	catalog<I32>("depth","max")=32;
	catalog<String>("depth","label")="Depth";
	catalog<String>("depth","hint")=
			"Number of subdivision iterations.";

	catalog<I32>("flatDepth")=0;
	catalog<I32>("flatDepth","high")=8;
	catalog<I32>("flatDepth","max")=32;
	catalog<String>("flatDepth","label")="Flat Depth";
	catalog<String>("flatDepth","hint")=
			"Number of iterations to linearly interpolate (at finest end).";

	catalog< sp<Component> >("Input Surface");

	//* don't copy input to output
	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";

	m_spSurfaceNormalOp=registry()->create("OperatorSurfaceI.SurfaceNormalOp");
}

void SubdivideOp::bind(sp<SignalerI> a_spSignalerI,sp<Layout> a_spLayout)
{
	OperatorSurfaceCommon::bind(a_spSignalerI,a_spLayout);

	sp<OperatorSurfaceCommon> spCommon=m_spSurfaceNormalOp;
	if(spCommon.isValid())
	{
		spCommon->bind(a_spSignalerI,a_spLayout);
	}
}

void SubdivideOp::handle(Record& a_rSignal)
{
	const String partitionAttr=catalog<String>("partitionAttr");
	const BWORD recomputeNormals=catalog<bool>("recomputeNormals");
	const I32 depth=catalog<I32>("depth");
	const I32 flatDepth=catalog<I32>("flatDepth");

	sp<SurfaceAccessibleI> spOutputAccessible;
	if(!accessOutput(spOutputAccessible,a_rSignal)) return;

	if(depth<1)
	{
		sp<SurfaceAccessibleI> spInputAccessible;
		if(!access(spInputAccessible,"Input Surface")) return;

		spOutputAccessible->copy(spInputAccessible);
	}

	for(I32 iteration=0;iteration<depth;iteration++)
	{
		const BWORD flat=(iteration>=depth-flatDepth);

		sp<SurfaceAccessibleI> spInputAccessible;
		if(iteration)
		{
			//* copy prior output to a new input buffer
			spInputAccessible=registry()->create("SurfaceAccessibleI");

			sp<Scope> spScope=registry()->create("Scope");
			spInputAccessible->bind(spScope);

			spInputAccessible->copy(spOutputAccessible);

			//* vacate prior output
			spOutputAccessible->clear();
		}
		else
		{
			if(!access(spInputAccessible,"Input Surface")) return;
		}

		sp<SurfaceAccessorI> spOutputVertices;
		if(!access(spOutputVertices,spOutputAccessible,
				e_primitive,e_vertices)) return;

		sp<SurfaceAccessorI> spOutputPoint;
		if(!access(spOutputPoint,spOutputAccessible,
				e_point,e_position)) return;

		sp<SurfaceAccessorI> spOutputNormal;
		if(!access(spOutputNormal,spOutputAccessible,
				e_point,e_normal)) return;

		sp<SurfaceAccessorI> spOutputUV;
		if(!access(spOutputUV,spOutputAccessible,
				e_point,"uv")) return;

		sp<SurfaceAccessorI> spOutputPartition;
		if(!access(spOutputPartition,spOutputAccessible,
				e_primitive,partitionAttr)) return;

		sp<SurfaceAccessorI> spInputVertices;
		if(!access(spInputVertices,spInputAccessible,
				e_primitive,e_vertices)) return;

		sp<SurfaceAccessorI> spInputPoint;
		if(!access(spInputPoint,spInputAccessible,
				e_point,e_position)) return;

		sp<SurfaceAccessorI> spInputNormal;
		if(!access(spInputNormal,spInputAccessible,
				e_point,e_normal)) return;

		sp<SurfaceAccessorI> spInputUV;
		if(!access(spInputUV,spInputAccessible,
				e_point,"uv")) return;

		sp<SurfaceAccessorI> spInputPartition;
		if(iteration && !access(spInputPartition,spInputAccessible,
				e_primitive,partitionAttr)) return;

		//* Catmull-Clark
		const U32 primitiveCountIn=spInputVertices->count();
		const U32 pointCountIn=spInputPoint->count();

#if FE_SDO_DEBUG
		feLog("SubdivideOp::handle input points %d primitives %d\n",
				pointCountIn,primitiveCountIn);
#endif

		std::vector<I32> facePoint(primitiveCountIn);

		std::vector<SpatialVector> faceSum(pointCountIn,
				SpatialVector(0.0,0.0,0.0));
		std::vector<I32> faceCount(pointCountIn,0);

		std::vector<SpatialVector> midpointSum(pointCountIn,
				SpatialVector(0.0,0.0,0.0));
		std::vector<I32> midpointCount(pointCountIn,0);

		std::vector<SpatialVector> boundarySum(pointCountIn,
				SpatialVector(0.0,0.0,0.0));
		std::vector<I32> boundaryCount(pointCountIn,0);

		//* face list per edge
		std::map<I32, std::map<I32,Edge> > edgeMap;

		//* make slots for original points (with adjustments)
		for(U32 pointIndexIn=0;pointIndexIn<pointCountIn;pointIndexIn++)
		{
			spOutputPoint->append();
			spOutputNormal->set(pointIndexIn,
					spInputNormal->spatialVector(pointIndexIn));
			spOutputUV->set(pointIndexIn,
					spInputUV->spatialVector(pointIndexIn));
		}

		for(U32 primitiveIndexIn=0;primitiveIndexIn<primitiveCountIn;
				primitiveIndexIn++)
		{
			String partition;
			if(iteration)
			{
				partition=spInputPartition->string(primitiveIndexIn);
			}
			else
			{
				partition.sPrintf("%d",primitiveIndexIn);
			}

			SpatialVector pointSum(0.0,0.0,0.0);
			SpatialVector normalSum(0.0,0.0,0.0);
			SpatialVector uvSum(0.0,0.0,0.0);

			//* UV wrap-around
			Vector2i highUV(false,false);
			SpatialVector lowUV(0.0,0.0,0.0);

			const U32 subCount=spInputVertices->subCount(primitiveIndexIn);
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const I32 pointIndexIn=
						spInputVertices->integer(primitiveIndexIn,subIndex);
				const SpatialVector uv=spInputUV->spatialVector(pointIndexIn);

				pointSum+=spInputPoint->spatialVector(pointIndexIn);
				normalSum+=spInputNormal->spatialVector(pointIndexIn);
				uvSum+=uv;

				if(uv[0]>0.7)
				{
					highUV[0]=true;
				}
				if(uv[1]>0.7)
				{
					highUV[1]=true;
				}
				lowUV[0]+=(uv[0]<0.3);
				lowUV[1]+=(uv[1]<0.3);
			}

			if(highUV[0])
			{
				uvSum[0]+=lowUV[0];
			}
			if(highUV[1])
			{
				uvSum[1]+=lowUV[1];
			}

			if(subCount)
			{
				const Real scalar=1.0/Real(subCount);

				pointSum*=scalar;
				uvSum*=scalar;

				normalSum=unitSafe(normalSum);
			}

			if(uvSum[0]>=1.0)
			{
				uvSum[0]-=1.0;
			}
			if(uvSum[0]>=1.0)
			{
				uvSum[0]-=1.0;
			}

			//* face point
			const I32 pointIndexOut=spOutputPoint->append();

			spOutputPoint->set(pointIndexOut,pointSum);
			spOutputNormal->set(pointIndexOut,normalSum);
			spOutputUV->set(pointIndexOut,uvSum);

			facePoint[primitiveIndexIn]=pointIndexOut;

			I32 firstPrimitiveIndexOut= -1;
			I32 lastEdgePointIndex= -1;
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				//* edge face
				const I32 pointIndexIn=spInputVertices->integer(
						primitiveIndexIn,subIndex);
				const I32 nextPointIndexIn=spInputVertices->integer(
						primitiveIndexIn,(subIndex+1)%subCount);

				Edge& rEdge=(pointIndexIn>nextPointIndexIn)?
						edgeMap[pointIndexIn][nextPointIndexIn]:
						edgeMap[nextPointIndexIn][pointIndexIn];

				rEdge.m_faceSet.insert(primitiveIndexIn);
				if(rEdge.m_pointIndex<0)
				{
					rEdge.m_pointIndex=spOutputPoint->append();
				}

				//* face influence on pre-existing points
				faceSum[pointIndexIn]+=pointSum;
				faceCount[pointIndexIn]++;

				//* create sub primitive
				const I32 primitiveIndexOut=spOutputVertices->append();

				spOutputPartition->set(primitiveIndexOut,partition);

				//* vertex 0: original point
				spOutputVertices->append(primitiveIndexOut,pointIndexIn);

				//* vertex 1: midpoint
				spOutputVertices->append(primitiveIndexOut,rEdge.m_pointIndex);

				//* vertex 2: face point
				spOutputVertices->append(primitiveIndexOut,pointIndexOut);

				//* vertex 3: prev midpoint
				spOutputVertices->append(primitiveIndexOut,lastEdgePointIndex);

				if(subIndex==subCount-1)
				{
					spOutputVertices->set(firstPrimitiveIndexOut,3,
							rEdge.m_pointIndex);
				}

				if(!subIndex)
				{
					firstPrimitiveIndexOut=primitiveIndexOut;
				}
				lastEdgePointIndex=rEdge.m_pointIndex;
			}
		}

		//* set edge points
		for(std::map<I32, std::map<I32,Edge> >::iterator itA=edgeMap.begin();
				itA!=edgeMap.end();itA++)
		{
			const I32 pointIndexIn=itA->first;
			std::map<I32,Edge>& rEdges=itA->second;

			const SpatialVector pointA=
					spInputPoint->spatialVector(pointIndexIn);
			const SpatialVector normalA=
					spInputNormal->spatialVector(pointIndexIn);
			const SpatialVector uvA=
					spInputUV->spatialVector(pointIndexIn);

			for(std::map<I32,Edge>::iterator itB=rEdges.begin();
					itB!=rEdges.end();itB++)
			{
				const I32 nextPointIndexIn=itB->first;
				const Edge& rEdge=itB->second;
				const std::set<I32>& rFaceSet=rEdge.m_faceSet;

				const SpatialVector pointB=
						spInputPoint->spatialVector(nextPointIndexIn);
				const SpatialVector normalB=
						spInputNormal->spatialVector(nextPointIndexIn);
				const SpatialVector uvB=
						spInputUV->spatialVector(nextPointIndexIn);

				SpatialVector pointSum=pointA+pointB;
				SpatialVector normalSum=normalA+normalB;
				SpatialVector uvSum=uvA+uvB;

				const SpatialVector midpoint=0.5*pointSum;
				I32 divisor=2;

				//* UV wrap-around
				Vector2i highUV(false,false);
				SpatialVector lowUV(0.0,0.0,0.0);
				if(uvA[0]>0.7 || uvB[0]>0.7)
				{
					highUV[0]=true;
				}
				if(uvA[1]>0.7 || uvB[1]>0.7)
				{
					highUV[1]=true;
				}
				lowUV[0]+=(uvA[0]<0.3)+(uvB[0]<0.3);
				lowUV[1]+=(uvA[1]<0.3)+(uvB[1]<0.3);

				if(!flat && rFaceSet.size()>1)
				{
					//* also add adjacent face points
					for(std::set<I32>::iterator it=rFaceSet.begin();
							it!=rFaceSet.end();it++)
					{
						const I32 primitiveIndexIn=*it;
						const I32 pointIndexOut=facePoint[primitiveIndexIn];
						const SpatialVector uv=
								spOutputUV->spatialVector(pointIndexOut);

						pointSum+=spOutputPoint->spatialVector(pointIndexOut);
						normalSum+=spOutputNormal->spatialVector(pointIndexOut);
						uvSum+=uv;
						divisor++;

						if(uv[0]>0.7)
						{
							highUV[0]=true;
						}
						if(uv[1]>0.7)
						{
							highUV[1]=true;
						}
						lowUV[0]+=(uv[0]<0.3);
						lowUV[1]+=(uv[1]<0.3);
					}

					midpointSum[pointIndexIn]+=midpoint;
					midpointCount[pointIndexIn]++;

					midpointSum[nextPointIndexIn]+=midpoint;
					midpointCount[nextPointIndexIn]++;
				}
				else
				{
					boundarySum[pointIndexIn]+=midpoint;
					boundaryCount[pointIndexIn]++;

					boundarySum[nextPointIndexIn]+=midpoint;
					boundaryCount[nextPointIndexIn]++;
				}

				if(highUV[0])
				{
					uvSum[0]+=lowUV[0];
				}
				if(highUV[1])
				{
					uvSum[1]+=lowUV[1];
				}

				const Real scalar=1.0/Real(divisor);

				pointSum*=scalar;
				uvSum*=scalar;

				normalSum=unitSafe(normalSum);

				if(uvSum[0]>=1.0)
				{
					uvSum[0]-=1.0;
				}
				if(uvSum[0]>=1.0)
				{
					uvSum[0]-=1.0;
				}

				spOutputPoint->set(rEdge.m_pointIndex,pointSum);
				spOutputNormal->set(rEdge.m_pointIndex,normalSum);
				spOutputUV->set(rEdge.m_pointIndex,uvSum);
			}
		}

		//* adjust original points
		if(flat)
		{
			for(U32 pointIndexIn=0;pointIndexIn<pointCountIn;pointIndexIn++)
			{
				spOutputPoint->set(pointIndexIn,
						spInputPoint->spatialVector(pointIndexIn));
			}
		}
		else
		{
			for(U32 pointIndexIn=0;pointIndexIn<pointCountIn;pointIndexIn++)
			{
				const SpatialVector pointIn=
						spInputPoint->spatialVector(pointIndexIn);

				const I32 pointFaceCount=faceCount[pointIndexIn];
				FEASSERT(pointFaceCount>0);

				const SpatialVector faceAverage=
						faceSum[pointIndexIn]/Real(pointFaceCount);

				if(pointFaceCount>2)
				{
					const I32 divisor=midpointCount[pointIndexIn]+
							boundaryCount[pointIndexIn];
					FEASSERT(divisor>0);
					const SpatialVector midpointAverage=
							(midpointSum[pointIndexIn]+
							boundarySum[pointIndexIn])/
							Real(divisor);

					const SpatialVector pointOut=
							(faceAverage+2.0*midpointAverage+
							(pointFaceCount-3)*pointIn)/Real(pointFaceCount);

					spOutputPoint->set(pointIndexIn,pointOut);
				}
				else
				{
					//* TODO find out what everyone else does on boundaries
					FEASSERT(boundaryCount[pointIndexIn]>0);
					const SpatialVector boundaryAverage=
							boundarySum[pointIndexIn]/
							Real(boundaryCount[pointIndexIn]);

					const I32 n=4;

					const SpatialVector pointOut=
							0.5*(boundaryAverage+pointIn);

					spOutputPoint->set(pointIndexIn,pointOut);
				}
			}
		}

#if FE_SDO_DEBUG
		feLog("SubdivideOp::handle"
				" iteration %d subdivided points %d primitives %d\n",
				iteration,spOutputPoint->count(),spOutputVertices->count());
#endif
	}

	if(recomputeNormals)
	{
		m_spSurfaceNormalOp->handle(a_rSignal);
	}
}

} // namespace
